# (C) Copyright 2023 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#
import itertools
import logging
import math
import re

LOG = logging.getLogger(__name__)


def coords_to_index(coords, shape) -> int:
    a = 0
    n = 1
    i = len(coords) - 1
    while i >= 0:
        a += coords[i] * n
        n *= shape[i]
        i -= 1
    return a


def index_to_coords(index: int, shape):
    assert isinstance(index, int), (index, type(index))

    result = [None] * len(shape)
    i = len(shape) - 1

    while i >= 0:
        result[i] = index % shape[i]
        index = index // shape[i]
        i -= 1

    result = tuple(result)

    assert len(result) == len(shape)
    return result


class FieldCube:
    def __init__(
        self,
        ds,
        *args,
        remapping=None,
        flatten_values=False,
    ):
        assert len(ds), f"No data in {ds}"

        self.flatten_values = flatten_values

        if len(args) == 1 and isinstance(args[0], (list, tuple)):
            args = args[0]

        names = []
        for a in args:
            if isinstance(a, str):
                names.append(a)
            elif isinstance(a, dict):
                names += list(a.keys())

        self._field_shape = None

        # Sort the source according to their
        # internal_args = reduce(operator.add, [Separator.split(a) for a in args], [])
        # for i in ds:
        #   print(i)
        # print("before")
        # print(ds[0])
        # print(ds[1])
        # print(ds[2])
        # print(ds[3])
        self.source = ds.order_by(*args, remapping=remapping)
        del ds
        # print("after")
        # print(self.source[0])
        # print(self.source[1])
        # print(self.source[2])
        # print(self.source[3])

        # Get a mapping of user names to unique values
        # With possible reduce dimentionality if the user use 'level+param'
        self.user_coords = self.source.unique_values(*names, remapping=remapping)

        self.user_shape = tuple(len(v) for k, v in self.user_coords.items())

        if math.prod(self.user_shape) != len(self.source):
            details = []
            for key, v in self.user_coords.items():
                details.append(f"{key=} ({len(v)}) {v}")
            assert not isinstance(
                self.source, str
            ), f"Not expecting a str here ({self.source})"
            for i, f in enumerate(self.source):
                details.append(f"{i}={f} {f.metadata('number')}")
                if i > 30:
                    details.append("...")
                    break

            msg = (
                f"Shape {self.user_shape} [{math.prod(self.user_shape):,}]"
                + f" does not match number of available fields {len(self.source):,}. "
                + f"Difference: {len(self.source)-math.prod(self.user_shape):,}\n"
                + "\n".join(details)
            )
            raise ValueError(msg)

    @property
    def field_shape(self):
        if self._field_shape is None:
            self._field_shape = self.source[0].shape

            if self.flatten_values:
                self._field_shape = (math.prod(self._field_shape),)

            assert isinstance(self._field_shape, tuple), (
                self._field_shape,
                self.source[0],
            )
        return self._field_shape

    @property
    def extended_user_shape(self):
        return self.user_shape + self.field_shape

    def __str__(self):
        content = ", ".join([f"{k}:{len(v)}" for k, v in self.user_coords.items()])
        return f"{self.__class__.__name__}({content} ({len(self.source)} fields))"

    def squeeze(self):
        # TODO: remove an dimensions of one.
        return self

    def __getitem__(self, indexes):
        # Make sure the requested indexes are a list of slices matching the shape

        if isinstance(indexes, int):
            indexes = (indexes,)  # make tuple

        if isinstance(indexes, list):
            indexes = tuple(indexes)

        assert isinstance(indexes, tuple), (type(indexes), indexes)

        indexes = list(indexes)

        if indexes[-1] is Ellipsis:
            indexes.pop()

        while len(indexes) < len(self.user_shape):
            indexes.append(slice(None, None, None))

        # Map the slices to a list of indexes per dimension
        coords = []
        for s, c in zip(indexes, self.user_coords.values()):
            lst = list(range(len(c)))[s]
            if not isinstance(lst, list):
                lst = [lst]
            coords.append(lst)

        # Transform the coordinates to a list of indexes for the underlying dataset
        dataset_indexes = []
        user_shape = self.user_shape
        for x in itertools.product(*coords):
            i = coords_to_index(x, user_shape)
            assert isinstance(i, int), i
            dataset_indexes.append(i)

        ds = self.source[tuple(dataset_indexes)]

        # If we match just one element, we return it
        if all(len(_) == 1 for _ in coords):
            return ds[0]

        # For more than one element, we return a new cube
        return FieldCube(
            ds,
            *self.user_coords,
            flatten_values=self.flatten_values,
        )

    def to_numpy(self, **kwargs):
        return self.source.to_numpy(**kwargs).reshape(*self.extended_user_shape)

    def _names(self, coords, reading_chunks=None, **kwargs):
        if reading_chunks is None:
            reading_chunks = list(coords.keys())
        if isinstance(reading_chunks, (list, tuple)):
            assert all(isinstance(_, str) for _ in reading_chunks)
            reading_chunks = {k: len(coords[k]) for k in reading_chunks}

        for k, requested in reading_chunks.items():
            full_len = len(coords[k])
            assert full_len == requested, "only full chunks supported for now"

        names = list(coords[a] for a, _ in reading_chunks.items())

        return names

    def count(self, reading_chunks=None):
        names = self._names(reading_chunks=reading_chunks, coords=self.user_coords)
        return math.prod(len(lst) for lst in names)

    def iterate_cubelets(self, reading_chunks=None, **kwargs):
        names = self._names(reading_chunks=reading_chunks, coords=self.user_coords)
        indexes = list(range(0, len(lst)) for lst in names)

        return (
            Cubelet(self, i, coords_names=n)
            for n, i in zip(itertools.product(*names), itertools.product(*indexes))
        )

    def chunking(self, chunks):
        if isinstance(chunks, (str, int)):
            m = re.match(r"(\d+)\s*(.*)?", str(chunks))
            if not m:
                raise ValueError(f"Cannot parse {chunks}")
            size = int(m.group(1))
            unit = m.group(2).lower()
            if unit == "":
                unit = "m"
            size *= dict(k=1024, m=1024 * 1024, g=1024 * 1024 * 1024)[unit[0]]

            # TODO: use dtype.
            # We assuming float32 = 4 bytes per

            rest = math.prod(self.extended_user_shape[1:]) * 4
            first = int(size / rest + 0.5)
            return (first,) + self.extended_user_shape[1:]

        if not chunks:
            return True  # Let ZARR choose

        lst = list(self.user_shape)
        for i, name in enumerate(self.user_coords):
            lst[i] = chunks.get(name, lst[i])

        return tuple(lst)


class Cubelet:
    def __init__(self, cube, coords, coords_names=None):
        self._coords_names = coords_names  # only for display purposes
        self.owner = cube
        assert all(isinstance(_, int) for _ in coords), coords
        self.coords = coords
        self.flatten_values = cube.flatten_values

    def __repr__(self):
        return (
            f"{self.__class__.__name__}({self.coords},index_names={self._coords_names})"
        )

    @property
    def extended_icoords(self):
        return self.coords

    def to_numpy(self, **kwargs):
        return self.owner[self.coords].to_numpy(
            reshape=not self.flatten_values, **kwargs
        )



# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import logging

from . import Reader

LOG = logging.getLogger(__name__)


class ODBReader(Reader):
    def to_pandas(self, **kwargs):
        try:
            import codc as odc
        except Exception:
            import pyodc as odc

            LOG.debug("Using pure Python odc decoder.")

        odc_read_odb_kwargs = kwargs.get("odc_read_odb_kwargs", {})
        return odc.read_odb(self.path, single=True, *odc_read_odb_kwargs)


def reader(source, path, magic=None, deeper_check=False):
    if magic is None or magic[:5] == b"\xff\xffODA":
        return ODBReader(source, path)



# (C) Copyright 2023 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import logging
from collections import namedtuple

from climetlab.core.constants import DATETIME
from climetlab.core.order import normalize_order_by
from climetlab.core.select import normalize_selection
from climetlab.decorators import cached_method, normalize, normalize_grib_key_values
from climetlab.indexing.database.sql import (
    SqlDatabase,
    SqlOrder,
    SqlRemapping,
    SqlSelection,
)
from climetlab.loaders import build_remapping
from climetlab.readers.grib.index.db import FieldsetInFilesWithDBIndex
from climetlab.utils.serialise import register_serialisation

LOG = logging.getLogger(__name__)

SqlResultCache = namedtuple("SqlResultCache", ["first", "length", "result"])


@normalize(DATETIME, "date-list", format="%Y-%m-%d %H:%M:%S")
def _normalize_grib_kwargs_values(**kwargs):
    return kwargs


class FieldsetInFilesWithSqlIndex(FieldsetInFilesWithDBIndex):
    DBCLASS = SqlDatabase
    DB_CACHE_SIZE = 100_000
    DB_DICT_CACHE_SIZE = 100_000

    def apply_filters(self, filters):
        obj = self
        for f in filters:
            obj = obj.filter(f)
        return obj

    def _find_all_coords_dict(self):
        return self.db._find_all_coords_dict()

    def unique_values(self, *coords, remapping=None, progress_bar=None):
        """
        Given a list of metadata attributes, such as date, param, levels,
        returns the list of unique values for each attributes
        """
        keys = coords

        remapping = build_remapping(remapping)
        print("Not using remapping here")

        coords = {k: None for k in coords}
        coords = self._normalize_kwargs_names(**coords)
        coords = list(coords.keys())
        print("coords:", coords)
        values = self.db.unique_values(*coords, remapping=remapping).values()

        dic = {k: v for k, v in zip(keys, values)}
        return dic

    def filter(self, filter):
        db = self.db.filter(filter)
        return self.__class__(db=db)

    def sel(self, *args, remapping=None, **kwargs):
        kwargs = normalize_selection(*args, **kwargs)
        kwargs = self._normalize_kwargs_names(**kwargs)
        kwargs = normalize_grib_key_values(kwargs, as_tuple=True)
        if DATETIME in kwargs and kwargs[DATETIME] is not None:
            kwargs = _normalize_grib_kwargs_values(**kwargs)

        if not kwargs:
            return self
        return self.filter(SqlSelection(kwargs, remapping))

    def order_by(self, *args, remapping=None, **kwargs):
        kwargs = normalize_order_by(*args, **kwargs)
        kwargs = self._normalize_kwargs_names(**kwargs)

        out = self

        if remapping is not None:
            out = out.filter(SqlRemapping(remapping=remapping))

        if kwargs:
            out = out.filter(SqlOrder(kwargs))

        return out

    def part(self, n):
        if self._cache is None or not (
            self._cache.first <= n < self._cache.first + self._cache.length
        ):
            first = (n // self.DB_CACHE_SIZE) * self.DB_CACHE_SIZE
            result = self.db.lookup_parts(limit=self.DB_CACHE_SIZE, offset=first)
            self._cache = SqlResultCache(first, len(result), result)
        return self._cache.result[n % self.DB_CACHE_SIZE]

    def get_metadata(self, n):
        if self._dict_cache is None or not (
            self._dict_cache.first
            <= n
            < self._dict_cache.first + self._dict_cache.length
        ):
            first = (n // self.DB_DICT_CACHE_SIZE) * self.DB_DICT_CACHE_SIZE
            result = self.db.lookup_dicts(
                limit=self.DB_DICT_CACHE_SIZE,
                offset=first,
                with_parts=False,
                # remove_none=False ?
            )
            result = list(result)

            self._dict_cache = SqlResultCache(first, len(result), result)
        return self._dict_cache.result[n % self.DB_DICT_CACHE_SIZE]

    @cached_method
    def number_of_parts(self):
        return self.db.count()


register_serialisation(
    FieldsetInFilesWithSqlIndex,
    lambda x: [x.db.db_path, x.db._filters],
    lambda x: FieldsetInFilesWithSqlIndex(db=SqlDatabase(x[0])).apply_filters(
        filters=x[1]
    ),
)



# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import logging

import xarray as xr
from xarray.backends.common import BackendEntrypoint

LOG = logging.getLogger(__name__)


# We wrap the sources because the FileSource is a os.PathLike and
# since version 0.20, xarray checks the class and change os.PathLike to
# strings. We don't want that, as we want to keep our objects
class WrappedSource:
    def __init__(self, source):
        self.source = source


class CMLEngine(BackendEntrypoint):
    @classmethod
    def open_dataset(cls, filename_or_obj, *args, **kwargs):
        assert isinstance(filename_or_obj, WrappedSource)
        return filename_or_obj.source.to_xarray()


def infer_open_mfdataset_kwargs(
    sources=None,
    paths=None,
    reader_class=None,
    user_kwargs={},
):
    result = {}
    result.update(user_kwargs.get("xarray_open_mfdataset_kwargs", {}))
    if False:
        ds = sources[0].to_xarray()
        # lat_dims = [s.get_lat_dim() for s in sources]

        if ds.dims == ["lat", "lon", "forecast_time"]:
            result["concat_dim"] = "forecast_time"

        result.update(user_kwargs)
    return result


def merge(
    sources=None,
    paths=None,
    reader_class=None,
    **kwargs,
):
    assert sources

    options = infer_open_mfdataset_kwargs(
        sources=sources,
        paths=paths,
        reader_class=reader_class,
        user_kwargs=kwargs,
    )

    if reader_class is not None and hasattr(
        reader_class, "to_xarray_multi_from_sources"
    ):
        return reader_class.to_xarray_multi_from_sources(
            sources,
            **options,
        )

    if paths is not None:
        if reader_class is not None and hasattr(
            reader_class, "to_xarray_multi_from_paths"
        ):
            return reader_class.to_xarray_multi_from_paths(
                paths,
                **options,
            )

        LOG.debug(f"xr.open_mfdataset with options={options}")
        return xr.open_mfdataset(paths, **options)

    LOG.debug(f"xr.open_mfdataset with options= {options}")
    return xr.open_mfdataset(
        [WrappedSource(s) for s in sources],
        engine=CMLEngine,
        **options,
    )



# (C) Copyright 2021 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#


import json

from termcolor import colored

from .tools import parse_args, print_table


class SettingsCmd:
    @parse_args(
        args=(None, dict(metavar="SETTING", type=str, nargs="*")),
        json=dict(action="store_true", help="produce a JSON output"),
    )
    def do_settings(self, args):
        """
        Display or change CliMetLab settings.
        See https://climetlab.readthedocs.io/guide/settings.html for more details.

        Examples: climetlab settings cache-directory /big-disk/climetlab-cache
        """
        from climetlab import settings

        words = args.args

        if len(words) == 0:
            if args.json:
                result = {}
                for f in settings.dump():
                    result[f[0]] = f[1]
                print(json.dumps(result, indent=4, sort_keys=True))
                return

            print_table((f[0], f[1]) for f in settings.dump())
            return

        if len(words) == 1:
            name = words[0]
            print(settings.get(name))
            return

        if len(words) == 2:
            name = words[0]
            value = words[1]
            settings.set(name, value)

        if len(words) == 3:
            name = words[0]
            key = words[1]
            value = words[2]
            settings.set(name, {key: value})

    def complete_settings(self, text, line, start_index, end_index):
        from climetlab import settings

        names = [f[0] for f in settings.dump()]
        return [t for t in names if t.startswith(text)]

    def complete_settings_reset(self, *args, **kwargs):
        return self.complete_settings(*args, **kwargs)

    @parse_args(
        all=dict(action="store_true", help="All settings"),
        args=("args", dict(metavar="SETTING", type=str, nargs="*")),
    )
    def do_settings_reset(self, args):
        """
        Display or change CliMetLab settings.
        See https://climetlab.readthedocs.io/guide/settings.html for more details.

        Examples: climetlab settings cache-directory /big-disk/climetlab-cache
        """
        from climetlab import settings

        words = args.args

        if len(words) == 0:
            if args.all:
                settings.reset()
                return

            print(
                colored(
                    "To reset all settings, please use the --all flag. Use --help for more information.",
                    "red",
                )
            )
            return

        if len(words) == 1:
            name = words[0]
            settings.reset(name)
            return

        print(
            colored(
                f"Too many settings to reset ({words}).",
                "red",
            )
        )



# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

""" CliMetLab plugins rely on the python plugins_ system using ``entry_points``.

.. _plugins: https://packaging.python.org/guides/creating-and-discovering-plugins/

"""

import logging
import os
import sys
from collections import defaultdict
from importlib import import_module
from typing import List, Union

import entrypoints

import climetlab
from climetlab import settings
from climetlab.utils.humanize import did_you_mean

LOG = logging.getLogger(__name__)

CACHE = {}

PLUGINS = {}

REGISTERED = defaultdict(dict)

AVAILABLE_KINDS = ["dataset", "source"]


def refresh(kind=None):
    if kind in PLUGINS:
        PLUGINS.pop(kind)
        return
    if kind is None:
        PLUGINS.clear()
        return
    assert kind in AVAILABLE_KINDS, (kind, AVAILABLE_KINDS)


def _load_plugins(kind):
    plugins = {}
    for e in entrypoints.get_group_all(f"climetlab.{kind}s"):
        plugins[e.name.replace("_", "-")] = e
    return plugins


def load_plugins(kind):
    """Loads the plugins for a given kind. The plugin needs to have registered itself with entry_point.

    Parameters
    ----------
    kind : str
        Plugin type such as "dataset" or "source".
    """
    if PLUGINS.get(kind) is None:
        PLUGINS[kind] = _load_plugins(kind)
    return PLUGINS[kind]


def find_plugin(directories: Union[str, List[str]], name: str, loader, refreshed=False):
    """Find a plugin by name .

    Parameters
    ----------
    directories : list or str
        List of directories to be searched to find the plugin.
    name : str
        Name of the plugin
    loader : class
        Class implementing load_yaml() and load_module()

    Returns
    -------
    Return what the loader will returns when applied to the plugin with the right name `name`, found in one of the directories of the `directories` list.

    Raises
    ------
    NameError
        If plugin is not found.
    """  # noqa: E501
    candidates = set()

    if name in REGISTERED[loader.kind]:
        return getattr(REGISTERED[loader.kind][name], loader.kind)

    candidates.update(REGISTERED[loader.kind].keys())

    plugins = load_plugins(loader.kind)

    if name in plugins:
        plugin = plugins[name]
        return loader.load_entry(plugin)

    candidates.update(plugins.keys())

    if not isinstance(directories, (tuple, list)):
        directories = [directories]

    for directory in directories:
        n = len(directory)
        for path, _, files in os.walk(directory):
            path = path[n:]
            for f in files:
                base, ext = os.path.splitext(f)
                if ext == ".yaml":
                    candidates.add(base)
                    if base == name:
                        full = os.path.join(directory, path, f)
                        return loader.load_yaml(full)

                if ext == ".py" and base[0] != "_":
                    full = os.path.join(path, base)

                    if sys.platform == "win32":
                        full = full.replace("\\", "/")

                    if full[0] != "/":
                        full = "/" + full

                    p = full[1:].replace("/", "-").replace("_", "-")
                    candidates.add(p)
                    if p == name:
                        return loader.load_module(full.replace("/", "."))

    if not refreshed:
        LOG.debug("Cannot find {loader.kind} 'name'. Refreshing plugin list.")
        refresh(loader.kind)
        return find_plugin(directories, name, loader, refreshed=True)

    module = loader.load_remote(name)
    if module is not None:
        return module

    correction = did_you_mean(name, candidates)

    if correction is not None:
        LOG.warning(
            "Cannot find %s '%s', did you mean '%s'?",
            loader.kind,
            name,
            correction,
        )

    candidates = ", ".join(sorted(c for c in candidates if "-" in c))

    raise NameError(f"Cannot find {loader.kind} '{name}' (values are: {candidates})")


def directories(owner: bool = False) -> list:
    """Return a list of directories that are used in the project .

    If owner = False, return a list of directories where to search for plugins.

    If owner = True, return a list of 2-uples to include the owner in the return value.

    Parameters
    ----------
    owner : bool, optional

    """

    result = []
    for conf in (
        "styles-directories",
        "projections-directories",
        "layers-directories",
        "datasets-directories",
        "dask-directories",
    ):
        for d in settings.get(conf):
            if os.path.exists(d) and os.path.isdir(d):
                result.append(("user-settings", d))

    for kind in ("dataset", "source"):
        for name, v in load_plugins(kind).items():
            try:
                module = import_module(v.module_name)
                result.append((name, os.path.dirname(module.__file__)))
            except Exception:
                LOG.error("Cannot load module %s", v.module_name, exc_info=True)

    result.append(("climetlab", os.path.dirname(climetlab.__file__)))

    if owner:
        return result

    return [x[1] for x in result]


class RegisteredPlugin:
    pass


def register(kind, name, proc):
    assert name not in REGISTERED[kind], (kind, name, REGISTERED)
    if not hasattr(proc, kind):
        o = RegisteredPlugin()
        setattr(o, kind, proc)
        proc = o
    REGISTERED[kind][name] = proc



# (C) Copyright 2022 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import logging

LOG = logging.getLogger(__name__)


def _find_device():
    try:
        import torch
    except ImportError:
        LOG.debug("Torch not found")
        return None

    if torch.backends.mps.is_available() and torch.backends.mps.is_built():
        return "mps"
    if torch.cuda.is_available() and torch.backends.cuda.is_built():
        return "cuda"
    LOG.debug("Found no GPU, using CPU")
    return "cpu"


device = _find_device()


def to_pytorch_dataloader(dataset, **kwargs):
    import torch

    default_kwargs = dict(
        batch_size=128,
        # multi-process data loading
        # use as many workers as you have cores on your machine
        num_workers=1,
        # default: no shuffle, so need to explicitly set it here
        shuffle=True,
        # uses pinned memory to speed up CPU-to-GPU data transfers
        # see https://pytorch.org/docs/stable/notes/cuda.html#cuda-memory-pinning
        pin_memory=True,
        # function used to collate samples into batches
        # if None then Pytorch uses the default collate_fn (see below)
        collate_fn=None,
    )
    merged_kwargs = {k: v for k, v in default_kwargs.items()}
    if kwargs:
        merged_kwargs.update(kwargs)

    return torch.utils.data.DataLoader(dataset, **merged_kwargs)



# (C) Copyright 2022 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import logging
import math
import os
from abc import abstractmethod

from climetlab.core.index import Index, MaskIndex, MultiIndex
from climetlab.decorators import normalize_grib_key_values, normalize_grib_keys
from climetlab.indexing.database import (
    FILEPARTS_KEY_NAMES,
    MORE_KEY_NAMES,
    MORE_KEY_NAMES_WITH_UNDERSCORE,
    STATISTICS_KEY_NAMES,
)
from climetlab.readers.grib.codes import GribField
from climetlab.readers.grib.fieldset import FieldSetMixin
from climetlab.utils import progress_bar
from climetlab.utils.availability import Availability

LOG = logging.getLogger(__name__)


class FieldSet(FieldSetMixin, Index):
    _availability = None

    def __init__(self, *args, **kwargs):
        if self.availability_path is not None and os.path.exists(
            self.availability_path
        ):
            self._availability = Availability(self.availability_path)

        Index.__init__(self, *args, **kwargs)

    @classmethod
    def new_mask_index(self, *args, **kwargs):
        return MaskFieldSet(*args, **kwargs)

    @property
    def availability_path(self):
        return None

    @classmethod
    def merge(cls, sources):
        assert all(isinstance(_, FieldSet) for _ in sources)
        return MultiFieldSet(sources)

    def available(self, request, as_list_of_dicts=False):
        from climetlab.utils.availability import Availability

        if not request:
            return None, None

        @normalize_grib_keys
        def f(**_):
            return _

        request = f(**request)
        request = normalize_grib_key_values(request, as_tuple=True)

        keys = list(request.keys())

        ds = self.sel(**request)

        def dicts():
            for i in progress_bar(
                iterable=range(len(ds)),
                desc="Building availability",
            ):
                metadata = ds.get_metadata(i)
                dic = {k: metadata.get(k, "-") for k in keys}
                dic = normalize_grib_key_values(dic, as_tuple=False)
                yield dic

        available = Availability(dicts())

        missing = available.missing(**request)

        if as_list_of_dicts:
            # available = available.as_list_of_dicts()
            # missing = missing.as_list_of_dicts()
            raise NotImplementedError()

        return dict(available=available, missing=missing)

    def _custom_availability(
        self, keys=None, ignore_keys=None, filter_keys=lambda k: True
    ):
        def dicts():
            for i in progress_bar(
                iterable=range(len(self)),
                desc="Building availability",
            ):
                if keys is not None:
                    dic = self.get_metadata(i)
                    dic = {k: str(dic.get(k, "-")) for k in keys}
                else:
                    dic = self.get_metadata(i)

                    for k in list(dic.keys()):
                        if not filter_keys(k):
                            dic.pop(k)
                            continue
                        if ignore_keys and k in ignore_keys:
                            dic.pop(k)
                            continue
                        if dic[k] is None:
                            dic.pop(k)
                            continue

                yield dic

        from climetlab.utils.availability import Availability

        return Availability(dicts())

    @property
    def availability(self):
        if self._availability is not None:
            return self._availability
        LOG.debug("Building availability")

        self._availability = self._custom_availability(
            ignore_keys=FILEPARTS_KEY_NAMES
            + STATISTICS_KEY_NAMES
            + MORE_KEY_NAMES_WITH_UNDERSCORE
            + MORE_KEY_NAMES
        )
        return self.availability

    def is_full_hypercube(self):
        non_empty_coords = {
            k: v
            for k, v in self.availability._tree.unique_values().items()
            if len(v) > 1
        }
        expected_size = math.prod([len(v) for k, v in non_empty_coords.items()])
        return len(self) == expected_size

    @normalize_grib_keys
    def _normalize_kwargs_names(self, **kwargs):
        return kwargs


class MaskFieldSet(FieldSet, MaskIndex):
    def __init__(self, *args, **kwargs):
        MaskIndex.__init__(self, *args, **kwargs)


class MultiFieldSet(FieldSet, MultiIndex):
    def __init__(self, *args, **kwargs):
        MultiIndex.__init__(self, *args, **kwargs)


class FieldSetInFiles(FieldSet):
    # Remote Fieldsets (with urls) are also here,
    # as the actual fieldset is accessed on a file in cache.
    # This class changes the interface (_getitem__ and __len__)
    # into the interface (part and number_of_parts).
    def _getitem(self, n):
        part = self.part(n)
        return GribField(part.path, part.offset, part.length)

    def __len__(self):
        return self.number_of_parts()

    @abstractmethod
    def part(self, n):
        self._not_implemented()

    @abstractmethod
    def number_of_parts(self):
        self._not_implemented()



# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import inspect
import logging
import os
from collections import defaultdict

import yaml

from climetlab.core.plugins import directories
from climetlab.decorators import locked
from climetlab.utils.html import css
from climetlab.utils.kwargs import merge_dicts

LOG = logging.getLogger(__name__)


YAML_FILES = None

IGNORE = [
    "magics.yaml",
    "colours.yaml",
    "conventions.yaml",
    "config.yaml",
    "units.yaml",
]


def _guess(data, path):
    if "areas" in data:
        return "domains"

    if "dataset" in data:
        return "datasets"

    if "magics" in data:
        if "msymb" in data["magics"]:
            return "styles"

        if "mcont" in data["magics"]:
            return "styles"

        if "mcoast" in data["magics"]:
            return "layers"

        if "mmap" in data["magics"]:
            return "projections"

        LOG.warning("Cannot guess collection for %s", path)

    if "dask" in data:
        return "dask"

    return "unknown"


class Entry:
    def __init__(self, name, kind, root, path, data, owner):
        self.name = name
        self.root = root
        self.kind = kind
        self.path = path
        self.data = data
        self.hidden = data.get("hidden", False)
        self.owner = owner
        self.next = None

    def _repr_html_(self):
        html = [
            css("table"),
            "<table class='climetlab'>",
            "<tr><td>Name:</td><td>%s</td></tr>" % self.name,
            "<tr><td>Collection:</td><td>%s</td></tr>" % self.kind,
            "<tr><td>Path:</td><td>%s</td></tr>" % self.path,
        ]

        html.append(
            "<tr><td>Definition:</td><td><pre>%s</pre></td></tr>"
            % (yaml.dump(self.data, default_flow_style=False),)
        )
        html.append("</table>")
        return "".join(html)

    def __repr__(self):
        return f"Entry({self.name}, path={self.path})"

    def choices(self):
        result = {self.owner: self}
        if self.next:
            result.update(self.next.choices())
        return result


@locked
def _load_yaml_files():
    global YAML_FILES
    if YAML_FILES is not None:
        return YAML_FILES

    YAML_FILES = defaultdict(dict)
    for owner, directory in directories(owner=True):
        for root, _, files in os.walk(directory):
            for file in [f for f in files if f.endswith(".yaml")]:
                if file in IGNORE:
                    continue
                path = os.path.join(root, file)
                try:
                    with open(path) as f:
                        data = yaml.load(f.read(), Loader=yaml.SafeLoader)
                        if not isinstance(data, dict):
                            continue
                        name, _ = os.path.splitext(os.path.basename(path))
                        kind = _guess(data, path)
                        collection = YAML_FILES[kind]
                        e = Entry(name, kind, directory, path, data, owner)
                        if name in collection:
                            e.next = collection[name]
                        collection[name] = e

                except Exception:
                    LOG.exception("Cannot process YAML file %s (%s)", path, owner)

    return YAML_FILES


def get_data_entry(kind, name, default=None, merge=False):
    """Read various config files to get the config values of
    a given kind.

    kind: Can be "layers", "projections", "dask", etc..
    name: The filename of the config file to read is something like {name}.yaml
    default: is a default dict when no config file is found
    merge: if False, only one config file will be used.
           if True, config files will be merged according to a given priority order.

    When looking for config files from a climetlab plugin, the config file provided
    by the plugin (if it exists) will be given higher priority than the file provided by
    climetlab.
    User-defined settings are always higher priority.
    """

    PRIORITIES = {
        # HIGHEST PRIORITY
        # "environment-variable": 6,
        "user-settings": 5,
        # "system-settings": 4,
        "plugins": 3,
        "climetlab": 2,
        "default": 1,
        # LOWEST PRIORITY
    }

    # TODO: default is not used. Remove it?
    # TODO: merge is not used.Remove it?

    files = _load_yaml_files()

    if kind not in files:
        if default is not None:
            return default
        raise KeyError("No collection named '%s'" % (kind,))

    if name not in files[kind]:
        if default is not None:
            return default
        raise KeyError(
            "No object '%s' in collection named '%s' (%s)"
            % (name, kind, sorted(files[kind].keys()))
        )

    choices = files[kind][name].choices()
    assert len(choices) != 0

    if len(choices) == 1:
        return list(choices.values())[0]

    frame = inspect.currentframe()
    caller = inspect.getouterframes(frame, 0)

    def is_active(owner, entry):
        # always active
        if owner in PRIORITIES:
            return True
        # check if called from the code inside a plugin
        for c in caller:
            if c.filename.startswith(entry.root):
                return True
        return False

    if default is not None:
        choices["default"] = Entry(
            name="default",
            kind="default",
            root=None,
            path=None,
            data=default,
            owner="default",
        )

    choices = {k: v for k, v in choices.items() if is_active(k, v)}
    selected = [
        v
        for _, v in sorted(
            choices.items(), key=lambda x: PRIORITIES.get(x[0], PRIORITIES["plugins"])
        )
    ]

    if not merge:
        return selected[0]

    data = merge_dicts(*[v.data for v in selected])
    return Entry(
        name="merged",
        kind=kind,
        root=None,
        path=None,
        data=data,
        owner="merged",
    )


def data_entries(kind=None):
    if kind is None:
        for collection in _load_yaml_files().items():
            for entry in collection.values():
                if not entry.hidden:
                    yield entry
    else:
        for entry in _load_yaml_files().get(kind, {}).values():
            if not entry.hidden:
                yield entry


@locked
def clear_cache():
    global YAML_FILES
    YAML_FILES = None



# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import logging
import mimetypes
import tarfile

from .archive import ArchiveReader

LOG = logging.getLogger(__name__)


class TarReader(ArchiveReader):
    def __init__(self, source, path, compression=None):
        super().__init__(source, path)

        with tarfile.open(path) as tar:
            self.expand(
                tar,
                tar.getmembers(),
                set_attrs=False,
            )


def reader(source, path, magic=None, deeper_check=False):
    # We don't use tarfile.is_tarfile() because is
    # returns true given a file of zeros

    kind, compression = mimetypes.guess_type(path)

    if magic is None or kind == "application/x-tar":
        return TarReader(source, path, compression)



# (C) Copyright 2021 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#
import calendar
import datetime
import re

from dateutil.parser import isoparse, parse

from climetlab.wrappers import Wrapper

VALID_DATE = re.compile(r"\d\d\d\d-?\d\d-?\d\d([T\s]\d\d:\d\d(:\d\d)?)?Z?")
MONTH_DATE = re.compile(r"^\d\d\d\d-\d\d?$")


def parse_date(dt):
    if not VALID_DATE.match(dt):
        raise ValueError(f"Invalid datetime '{dt}'")

    try:
        return datetime.datetime.fromisoformat(dt)
    except Exception:
        pass

    try:
        return isoparse(dt)
    except ValueError:
        pass

    return parse(dt)


class StrWrapper(Wrapper):
    def __init__(self, data):
        self.data = data

    def to_bounding_box(self):
        from climetlab.utils.domains import domain_to_area

        return domain_to_area(self.data)

    def to_datetime(self):
        return parse_date(self.data)

    def to_datetime_list(self):
        from climetlab.utils.dates import mars_like_date_list

        if MONTH_DATE.match(self.data):
            first_day = parse(self.data)
            y = first_day.year
            m = first_day.month
            n_days = calendar.monthrange(y, m)[1]
            return mars_like_date_list(
                parse_date(f"{y}-{m:02}-01"), parse_date(f"{y}-{m:02}-{n_days}"), 1
            )

        # MARS style lists
        bits = self.data.split("/")
        if len(bits) == 3 and bits[1].lower() == "to":
            return mars_like_date_list(parse_date(bits[0]), parse_date(bits[2]), 1)

        if len(bits) == 5 and bits[1].lower() == "to" and bits[3].lower() == "by":
            return mars_like_date_list(
                parse_date(bits[0]), parse_date(bits[2]), int(bits[4])
            )

        return [parse_date(d) for d in bits]


def wrapper(data, *args, **kwargs):
    if isinstance(data, str):
        return StrWrapper(data)
    return None



# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#
import inspect
import logging
import os
from urllib.parse import urlparse

from climetlab.sources.ecmwf_open_data import EODRetriever
from climetlab.sources.file import FileSource
from climetlab.sources.url import Url

from . import BaseMirror, MirrorConnection

LOG = logging.getLogger(__name__)


def strict_init(cls):
    sig = inspect.signature(cls.__init__)

    def check_type(name, value):
        type = sig.parameters[name].annotation
        if type == inspect._empty:
            return
        assert isinstance(value, type), (value, type)

    class Wrapped(cls):
        def __init__(self, mirror, source, *args, **kwargs):
            check_type("mirror", mirror)
            check_type("source", source)
            return super().__init__(mirror, source, *args, **kwargs)

    return Wrapped


class DirectoryMirror(BaseMirror):
    def __init__(self, path, origin_prefix=""):
        self.path = path
        self.origin_prefix = origin_prefix

    def __repr__(self):
        return f"DirectoryMirror({self.path}, {self.origin_prefix})"

    def connection_for_url(self, source, url, parts):
        if not url.startswith(self.origin_prefix):
            return None
        LOG.debug(f"Mirror for {url} activated.")
        if parts:
            LOG.debug(f"Mirror for {url} deactived for parts.")
            # return DirectoryMirrorConnectionForUrlWithPargs(self, source, url, parts)
            return None
        return DirectoryMirrorConnectionForUrl(self, source)

    def connection_for_eod(self, source):
        return DirectoryMirrorConnectionForEODRetriever(self, source)


class DirectoryMirrorConnection(MirrorConnection):
    def _realpath(self):
        keys = self._to_keys()
        assert isinstance(keys, (list, tuple)), type(keys)
        path = os.path.join(
            self.mirror.path,
            *keys,
        )
        return os.path.realpath(path)


class DirectoryMirrorConnectionForFile(DirectoryMirrorConnection):
    def __init__(self, mirror: DirectoryMirror, source: FileSource):
        assert isinstance(source, FileSource)
        return super().__init__(mirror, source)

    def resource(self):
        path = self._realpath()
        if os.path.exists(path):
            return path
        else:
            return None

    def create_copy(self, create, args):
        path = self._realpath()
        os.makedirs(os.path.dirname(path), exist_ok=True)
        create(path, args)
        return path


class DirectoryMirrorConnectionForEODRetriever(DirectoryMirrorConnectionForFile):
    def __init__(self, mirror: DirectoryMirror, source: EODRetriever):
        assert isinstance(source, EODRetriever)
        return super().__init__(mirror, source)

    def _to_keys(self):
        request = self.source.source_kwargs
        keys = []
        for k in request.keys():
            keys.append(k)
            keys.append(str(request[k]))
        return keys


class DirectoryMirrorConnectionForUrl(DirectoryMirrorConnectionForFile):
    def __init__(self, mirror: DirectoryMirror, source: Url):
        assert isinstance(source, Url)
        return super().__init__(mirror, source)

    def _to_keys(self):
        url = self.source.url
        if self.mirror.origin_prefix:
            key = url[(len(self.mirror.origin_prefix) + 1) :]
            return ["url", key]

        url = urlparse(url)
        keys = [url.scheme, f"{url.netloc}/{url.path}"]
        return ["url"] + keys



# (C) Copyright 2021 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import logging
import os

import climetlab as cml

from .tools import parse_args

LOG = logging.getLogger(__name__)


default_keys = ["param", "time", "date", "step", "levelist"]


class AvailabilityCmd:
    @parse_args(
        source=(
            None,
            dict(
                type=str,
                help="File or directory for describing a dataset or source of data with GRIB data.",
            ),
        ),
        stdout=dict(action="store_true", help="Output to stdout (no file)."),
        yaml=dict(action="store_true", help="Output yaml format."),
        keys=(
            "--keys",
            dict(
                type=str,
                help=f"Keys to use in availability default is ({','.join(default_keys)}).",
                nargs="+",
            ),
        ),
    )
    def do_availability(self, args):
        """Create json availability file."""

        keys = args.keys
        if keys == []:
            keys = default_keys

        if keys == ["*"]:
            keys = None

        self.avail = None

        path = args.source

        if not os.path.exists(path):
            print(f"{path} does not exists.")
            return

        if os.path.isdir(path):
            source = availability_of_directory(path)
        else:
            source = availability_of_file(path)

        availability = source._custom_availability(keys=keys)

        if args.yaml:
            print(availability.to_yaml())
            return

        if args.stdout:
            print(availability.tree())
            return

        # self.write(availability, source.availability_path)

    # def write(self, avail, output):
    #    if os.path.exists(output):
    #        i = 1
    #        while os.path.exists(f"{output}.{i}"):
    #            i += 1
    #        output = f"{output}.{i}"
    #        print(f"File already exists, writing to {output}")

    #    avail.to_pickle(output)


def availability_of_directory(dirpath):
    db_path = os.path.join(dirpath, "climetlab-2.db")
    if not os.path.exists(db_path):
        print(f"ERROR: this directory is not indexed yet, cannot find {db_path}.")

    source = cml.load_source("indexed-directory", dirpath)
    return source


def availability_of_file(path):
    # if magic = "SQLite"...
    return cml.load_source("file", path)



# (C) Copyright 2022 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import time

import numpy as np


class Helper:
    def __init__(self):
        self.best = None
        self.max = float("inf")


class KDNode:
    __slots__ = "value", "axis", "left", "right", "payload"

    def __init__(self, value, axis):
        self.value = value[:-1]
        self.payload = value[-1]
        self.axis = axis
        self.left = None
        self.right = None

    def visit(self, visitor, depth):
        visitor(self, depth)
        if self.left:
            self.left.visit(visitor, depth + 1)
        if self.right:
            self.right.visit(visitor, depth + 1)

    def find_nn(self, point):
        o = Helper()
        self._find_nn(point, o, 0)
        return (o.best, o.max)

    def _find_nn(self, point, o, depth):
        x = point[self.axis]
        y = self.value[self.axis]

        if x < y:
            if self.left:
                self.left._find_nn(point, o, depth + 1)
        else:
            if self.right:
                self.right._find_nn(point, o, depth + 1)

        d = np.linalg.norm(point - self.value)
        if d < o.max:
            o.max = d
            o.best = self.payload

        d = np.fabs(x - y)

        if d < o.max:
            # Cross over: Visit other subtree...
            if x < y:
                if self.right:
                    self.right._find_nn(point, o, depth + 1)
            else:
                if self.left:
                    self.left._find_nn(point, o, depth + 1)


class KDChunk:
    __slots__ = "values"

    def __init__(self, values):
        self.values = values

    def visit(self, visitor, depth):
        visitor(self, depth)

    def _find_nn(self, point, o, depth):
        d = min((np.linalg.norm(point - v[:-1]), v[-1]) for v in self.values)

        if d[0] < o.max:
            o.max = d[0]
            o.best = d[1]


class KDTree:
    def __init__(self, dimensions, values, chunk_size=-1):
        self.dimensions = dimensions
        self.chunk_size = chunk_size
        self.root = self.build(values)

    def build(self, values, depth=0):
        if len(values) == 0:
            return None

        if len(values) <= self.chunk_size:
            return KDChunk(values)

        k = self.dimensions
        axis = depth % k
        # print(values.shape, values[0])
        values = values[np.argsort(values[:, axis])]

        median = len(values) // 2

        node = KDNode(values[median], axis)
        node.left = self.build(values[:median], depth + 1)
        node.right = self.build(values[median + 1 :], depth + 1)

        return node

    def find_nn(self, point):
        return self.root.find_nn(point)

    def depth(self):
        class visitor:
            def __init__(self):
                self.depth = 0

            def __call__(self, _, depth):
                if depth > self.depth:
                    self.depth = depth

        v = visitor()
        self.root.visit(v, 0)
        return v.depth + 1

    def size(self):
        class visitor:
            def __init__(self):
                self.size = 0

            def __call__(self, _, depth):
                self.size += 1

        v = visitor()
        self.root.visit(v, 0)
        return v.size


CACHE = {}


def sin_cos(x):
    if x not in CACHE:
        y = np.deg2rad(x)
        CACHE[x] = (np.sin(y), np.cos(y))
    return CACHE[x]


def ecef(lat, lon, i):
    # ECEF (Earth-Centered, Earth-Fixed) coordinate system
    sin_lat, cos_lat = sin_cos(lat)
    sin_lon, cos_lon = sin_cos(lon)

    return np.array(
        [
            cos_lat * cos_lon,
            cos_lat * sin_lon,
            sin_lat,
            i,
        ]
    )


def unstructed_to_structed(grib, chunk_size=-1):
    now = time.time()
    print("----")
    xyz = np.array(
        [ecef(lat, lon, i) for i, (lat, lon) in enumerate(grib.iterate_grid_points())]
    )
    print("----", time.time() - now)
    print(len(xyz))

    now = time.time()
    tree = KDTree(3, xyz, chunk_size)
    print("----", time.time() - now)
    return tree


def lookup(tree, lat, lon):
    xyz = ecef(lat / 10, lon / 10, 0)
    return tree.find_nn(xyz[:-1])[0]
    # print(xyz)
    # print("----", time.time() - now)
    # print(tree.size(), tree.depth())
    # now = time.time()
    # z = []
    # for lat in range(900, -901, -1):
    #     for lon in range(0, 3601):
    #         xyz = ecef(lat / 10, lon / 10, 0)
    #         z.append(tree.find_nn(xyz[:-1])[0])
    # print("----", time.time() - now)
    # print(np.array(z))
    # return time.time() - start



# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import itertools
import re

from .dates import to_datetime

RE1 = re.compile(r"{([^}]*)}")
RE2 = re.compile(r"\(([^}]*)\)")

RE3 = re.compile(r"\(([^}]*)\)")


class Any:
    def substitute(self, value, name):
        return value


class Enum:
    def __init__(self, enum=""):
        self.enum = set(enum.split(","))

    def substitute(self, value, name):
        if self.enum and value not in self.enum:
            raise ValueError(
                "Invalid value '{}' for parameter '{}', expected one of {}".format(
                    value, name, self.enum
                )
            )
        return value


class Int:
    def __init__(self, format="%d"):
        self.format = format

    def substitute(self, value, name):
        if not isinstance(value, int):
            raise ValueError(
                "Invalid value '{}' for parameter '{}', expected an integer".format(
                    value, name
                )
            )
        return self.format % value


class Float:
    def __init__(self, format="%g"):
        self.format = format

    def substitute(self, value, name):
        if not isinstance(value, (int, float)):
            raise ValueError(
                "Invalid value '{}' for parameter '{}', expected a float".format(
                    value, name
                )
            )

        return self.format % value


class Datetime:
    def __init__(self, format):
        self.format = format

    def substitute(self, value, name):
        return to_datetime(value).strftime(self.format)


class Str:
    def __init__(self, format="%s"):
        self.format = format

    def substitute(self, value, name):
        if not isinstance(value, str):
            raise ValueError(
                "Invalid value '{}' for parameter '{}', expected a string".format(
                    value, name
                )
            )
        return self.format % value


TYPES = {"": Any, "int": Int, "float": Float, "date": Datetime, "enum": Enum}


class Constant:
    name = None

    def __init__(self, value):
        self.value = value

    def substitute(self, params):
        return self.value


class Variable:
    def __init__(self, value):
        bits = value.split(":")
        self.name = bits[0]
        kind = RE2.split(":".join(bits[1:]))
        if len(kind) == 1:
            self.kind = TYPES[kind[0]]()
        else:
            self.kind = TYPES[kind[0]](kind[1])

    def substitute(self, params):
        if self.name not in params:
            raise ValueError("Missing parameter '{}'".format(self.name))
        return self.kind.substitute(params[self.name], self.name)


FUNCTIONS = dict(lower=lambda s: s.lower())


class Function:
    def __init__(self, value):
        functions = value.split("|")
        self.name = functions[0]
        self.variable = Variable(functions[0])
        self.functions = functions[1:]

    def substitute(self, params):
        value = self.variable.substitute(params)
        for f in self.functions:
            value = FUNCTIONS[f](value)
        return value


class Pattern:
    def __init__(self, pattern, ignore_missing_keys=False):
        self.ignore_missing_keys = ignore_missing_keys

        self.pattern = []
        self.variables = []
        for i, p in enumerate(RE1.split(pattern)):
            if i % 2 == 0:
                self.pattern.append(Constant(p))
            else:
                if "|" in p:
                    v = Function(p)
                else:
                    v = Variable(p)
                self.variables.append(v)
                self.pattern.append(v)

    @property
    def names(self):
        return sorted({v.name for v in self.variables})

    def substitute(self, *args, **kwargs):
        params = {}
        for a in args:
            params.update(a)
        params.update(kwargs)

        for k, v in params.items():
            if isinstance(v, list):
                return self._substitute_many(params)

        return self._substitute_one(params)
        # TODO: discuss if this should be:
        # return [self._substitute_one(params)]

    def _substitute_one(self, params):
        used = set(params.keys())
        result = []
        for p in self.pattern:
            used.discard(p.name)
            result.append(p.substitute(params))
        if used and not self.ignore_missing_keys:
            raise ValueError("Unused parameter(s): {}".format(used))

        return "".join(str(x) for x in result)

    def _substitute_many(self, params):
        for k, v in list(params.items()):
            if not isinstance(v, list):
                params[k] = [v]

        seen = set()
        result = []
        for n in (
            dict(zip(params.keys(), x)) for x in itertools.product(*params.values())
        ):
            m = self.substitute(n)
            if m not in seen:
                seen.add(m)
                result.append(m)

        return result



# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import datetime
import logging

import numpy as np

from climetlab.wrappers import get_wrapper

LOG = logging.getLogger(__name__)


def to_datetime(dt):
    if isinstance(dt, datetime.datetime):
        return dt

    if isinstance(dt, datetime.date):
        return datetime.datetime(dt.year, dt.month, dt.day)

    if isinstance(dt, np.datetime64):
        # Looks like numpy dates conversion vary
        dt = dt.astype(datetime.datetime)

        if isinstance(dt, datetime.datetime):
            return dt

        if isinstance(dt, datetime.date):
            return to_datetime(dt)

        if isinstance(dt, int):
            return datetime.datetime.utcfromtimestamp(dt * 1e-9)

        raise ValueError("Failed to convert numpy datetime {}".format((dt, type(dt))))

    dt = get_wrapper(dt)

    return to_datetime(dt.to_datetime())


def mars_like_date_list(start, end, by):
    """Return a list of datetime objects from start to end .

    Parameters
    ----------
    start : [type]
        [description]
    end : [type]
        [description]
    by : [type]
        [description]

    Returns
    -------
    [type]
        [description]
    """
    assert by > 0, by
    assert end >= start
    result = []
    while start <= end:
        result.append(start)
        start = start + datetime.timedelta(days=by)
    return result


def to_datetime_list(datetimes):  # noqa C901
    if isinstance(datetimes, (datetime.datetime, np.datetime64)):
        return to_datetime_list([datetimes])

    if isinstance(datetimes, (list, tuple)):
        if (
            len(datetimes) == 3
            and isinstance(datetimes[1], str)
            and datetimes[1].lower() == "to"
        ):
            return mars_like_date_list(
                to_datetime(datetimes[0]), to_datetime(datetimes[2]), 1
            )

        if (
            len(datetimes) == 5
            and isinstance(datetimes[1], str)
            and isinstance(datetimes[3], str)
            and datetimes[1].lower() == "to"
            and datetimes[3].lower() == "by"
        ):
            return mars_like_date_list(
                to_datetime(datetimes[0]), to_datetime(datetimes[2]), int(datetimes[4])
            )

        return [to_datetime(x) for x in datetimes]

    datetimes = get_wrapper(datetimes)

    return to_datetime_list(datetimes.to_datetime_list())


def to_date_list(obj):
    return sorted(set(to_datetime_list(obj)))



# (C) Copyright 2021 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#
import logging

LOG = logging.getLogger(__name__)


def guess_type_list(lst):
    assert isinstance(lst, (tuple, list))

    for typ in [int, float, str]:
        if all([isinstance(x, typ) for x in lst]):
            return typ

    return None



# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import datetime
import logging
import os
import threading
import time
from itertools import islice

import eccodes

from climetlab.core import Base
from climetlab.core.constants import DATETIME
from climetlab.profiling import call_counter
from climetlab.utils.bbox import BoundingBox

LOG = logging.getLogger(__name__)


def missing_is_none(x):
    return None if x == 2147483647 else x


# This does not belong here, should be in the C library
def get_messages_positions(path):
    fd = os.open(path, os.O_RDONLY)
    try:

        def get(count):
            buf = os.read(fd, count)
            assert len(buf) == count
            return int.from_bytes(
                buf,
                byteorder="big",
                signed=False,
            )

        offset = 0
        while True:
            code = os.read(fd, 4)
            if len(code) < 4:
                break

            if code != b"GRIB":
                offset = os.lseek(fd, offset + 1, os.SEEK_SET)
                continue

            length = get(3)
            edition = get(1)

            if edition == 1:
                if length & 0x800000:
                    sec1len = get(3)
                    os.lseek(fd, 4, os.SEEK_CUR)
                    flags = get(1)
                    os.lseek(fd, sec1len - 8, os.SEEK_CUR)

                    if flags & (1 << 7):
                        sec2len = get(3)
                        os.lseek(fd, sec2len - 3, os.SEEK_CUR)

                    if flags & (1 << 6):
                        sec3len = get(3)
                        os.lseek(fd, sec3len - 3, os.SEEK_CUR)

                    sec4len = get(3)

                    if sec4len < 120:
                        length &= 0x7FFFFF
                        length *= 120
                        length -= sec4len
                        length += 4

            if edition == 2:
                length = get(8)

            yield offset, length
            offset = os.lseek(fd, offset + length, os.SEEK_SET)

    finally:
        os.close(fd)


eccodes_codes_release = call_counter(eccodes.codes_release)
eccodes_codes_new_from_file = call_counter(eccodes.codes_new_from_file)

# For some reason, cffi can ge stuck in the GC if that function
# needs to be called defined for the first time in a GC thread.
try:
    _h = eccodes.codes_new_from_samples(
        "regular_ll_pl_grib1", eccodes.CODES_PRODUCT_GRIB
    )
    eccodes.codes_release(_h)
except:  # noqa E722
    pass


class CodesHandle:
    def __init__(self, handle, path, offset):
        self.handle = handle
        self.path = path
        self.offset = offset

    @classmethod
    def from_sample(cls, name):
        return cls(
            eccodes.codes_new_from_samples(name, eccodes.CODES_PRODUCT_GRIB), None, None
        )

    def __del__(self):
        try:
            eccodes_codes_release(self.handle)
        except TypeError:
            # This happens when eccodes is unloaded before
            # this object is deleted
            pass

    def get(self, name):
        try:
            if name == "values":
                return eccodes.codes_get_values(self.handle)
            size = eccodes.codes_get_size(self.handle, name)
            # LOG.debug(f"{name}:{size}")

            if name == "md5GridSection":
                # Special case because:
                #
                # 1) eccodes is returning size > 1 for 'md5GridSection'
                # (size = 16 : it is the number of bytes of the value)
                # This will be fixed in eccodes.
                #
                # 2) sometimes (see below), the value for "shapeOfTheEarth" is inconsistent.
                # This impacts the (computed on-the-fly) value of "md5GridSection".
                # ----------------
                # Example of data with inconsistent values:
                # S2S data, origin='ecmf', param='tp', step='24', number='0', date=['20201203','20200702']
                # the 'md5GridSection' are different
                # This is because one has "shapeOfTheEarth" set to 0, the other to 6.
                # This is only impacting the metadata.
                # Since this has no impact on the data itself,
                # this is unlikely to be fixed. Therefore this hacky patch.
                #
                # Obviously, the patch causes an inconsistency between the value of md5GridSection
                # read by this code, and the value read by another code without this patch.

                save = eccodes.codes_get_long(self.handle, "shapeOfTheEarth")
                eccodes.codes_set_long(self.handle, "shapeOfTheEarth", 255)
                result = eccodes.codes_get_string(self.handle, "md5GridSection")
                eccodes.codes_set_long(self.handle, "shapeOfTheEarth", save)
                return result

            if size and size > 1:
                return eccodes.codes_get_array(self.handle, name)
            return eccodes.codes_get(self.handle, name)
        except eccodes.KeyValueNotFoundError:
            return None

    def get_long(self, name):
        try:
            return eccodes.codes_get_long(self.handle, name)
        except eccodes.KeyValueNotFoundError:
            return None

    def get_string(self, name):
        try:
            return eccodes.codes_get_string(self.handle, name)
        except eccodes.KeyValueNotFoundError:
            return None

    def get_double(self, name):
        try:
            return eccodes.codes_get_double(self.handle, name)
        except eccodes.KeyValueNotFoundError:
            return None

    def get_data(self):
        return eccodes.codes_grib_get_data(self.handle)

    def as_mars(self, param="shortName"):
        r = {}
        it = eccodes.codes_keys_iterator_new(self.handle, "mars")

        try:
            while eccodes.codes_keys_iterator_next(it):
                key = eccodes.codes_keys_iterator_get_name(it)
                r[key] = self.get(param if key == "param" else key)
        finally:
            eccodes.codes_keys_iterator_delete(it)

        return r

    def clone(self):
        return CodesHandle(eccodes.codes_clone(self.handle), None, None)

    def set_values(self, values):
        assert self.path is None, "Only cloned handles can have values changed"
        eccodes.codes_set_values(self.handle, values.flatten())
        # This is writing on the GRIB that something has been modified (255=unknown)
        eccodes.codes_set_long(self.handle, "generatingProcessIdentifier", 255)

    def set_multiple(self, values):
        assert self.path is None, "Only cloned handles can have values changed"
        eccodes.codes_set_key_vals(self.handle, values)

    def set_long(self, name, value):
        try:
            assert self.path is None, "Only cloned handles can have values changed"
            eccodes.codes_set_long(self.handle, name, value)
        except Exception as e:
            LOG.error("Error setting %s=%s", name, value)
            raise ValueError("Error setting %s=%s (%s)" % (name, value, e))

    def set_double(self, name, value):
        try:
            assert self.path is None, "Only cloned handles can have values changed"
            eccodes.codes_set_double(self.handle, name, value)
        except Exception as e:
            LOG.error("Error setting %s=%s", name, value)
            raise ValueError("Error setting %s=%s (%s)" % (name, value, e))

    def set_string(self, name, value):
        try:
            assert self.path is None, "Only cloned handles can have values changed"
            eccodes.codes_set_string(self.handle, name, value)
        except Exception as e:
            LOG.error("Error setting %s=%s", name, value)
            raise ValueError("Error setting %s=%s (%s)" % (name, value, e))

    def set(self, name, value):
        try:
            assert self.path is None, "Only cloned handles can have values changed"

            if isinstance(value, list):
                return eccodes.codes_set_array(self.handle, name, value)

            return eccodes.codes_set(self.handle, name, value)
        except Exception as e:
            LOG.error("Error setting %s=%s", name, value)
            raise ValueError("Error setting %s=%s (%s)" % (name, value, e))

    def write(self, f):
        eccodes.codes_write(self.handle, f)

    def save(self, path):
        with open(path, "wb") as f:
            eccodes.codes_write(self.handle, f)
            self.path = path
            self.offset = 0

    def read_bytes(self, offset, length):
        with open(self.path, "rb") as f:
            f.seek(offset)
            return f.read(length)


class ReaderLRUCache(dict):
    def __init__(self, size):
        self.readers = dict()
        self.lock = threading.Lock()
        self.size = size

    def __getitem__(self, path):
        key = (
            path,
            os.getpid(),
        )  # PyTorch will fork and leave the cache in a funny state
        with self.lock:
            try:
                return super().__getitem__(key)
            except KeyError:
                pass

            c = self[key] = CodesReader(path)

            while len(self) >= self.size:
                _, oldest = min((v.last, k) for k, v in self.items())
                del self[oldest]

            return c


cache = ReaderLRUCache(512)  # TODO: Add to config


class CodesReader:
    def __init__(self, path):
        self.path = path
        self.lock = threading.Lock()
        self.file = open(self.path, "rb")
        self.last = time.time()

    def __del__(self):
        try:
            self.file.close()
        except Exception:
            pass

    @classmethod
    def from_cache(cls, path):
        return cache[path]

    def at_offset(self, offset):
        with self.lock:
            self.last = time.time()
            self.file.seek(offset, 0)
            handle = eccodes_codes_new_from_file(
                self.file,
                eccodes.CODES_PRODUCT_GRIB,
            )
            assert handle is not None, (self.file, offset)
            return CodesHandle(handle, self.path, offset)


class GribField(Base):
    def __init__(self, path, offset, length):
        self.path = path
        self._offset = offset
        self._length = length
        self._handle = None
        self._values = None
        self._cache = {}

    @property
    def handle(self):
        if self._handle is None:
            assert self._offset is not None
            self._handle = CodesReader.from_cache(self.path).at_offset(self._offset)
        return self._handle

    @property
    def values(self):
        if self._values is None:
            self._values = self.handle.get("values")
        return self._values

    @property
    def offset(self):
        if self._offset is None:
            self._offset = int(self.handle.get("offset"))
        return self._offset

    @property
    def shape(self):
        Nj = missing_is_none(self.handle.get("Nj"))
        Ni = missing_is_none(self.handle.get("Ni"))
        if Ni is None or Nj is None:
            n = self.handle.get("numberOfDataPoints")
            return (n,)  # shape must be a tuple
        return (Nj, Ni)

    @property
    def mars_grid(self):
        if len(self.shape) == 2:
            return [
                self.handle.get("iDirectionIncrementInDegrees"),
                self.handle.get("jDirectionIncrementInDegrees"),
            ]

        return self.handle.get("gridName")

    @property
    def mars_area(self):
        north = self.handle.get("latitudeOfFirstGridPointInDegrees")
        south = self.handle.get("latitudeOfLastGridPointInDegrees")
        west = self.handle.get("longitudeOfFirstGridPointInDegrees")
        east = self.handle.get("longitudeOfLastGridPointInDegrees")
        return [north, west, south, east]

    def plot_map(self, backend):
        backend.bounding_box(
            north=self.handle.get("latitudeOfFirstGridPointInDegrees"),
            south=self.handle.get("latitudeOfLastGridPointInDegrees"),
            west=self.handle.get("longitudeOfFirstGridPointInDegrees"),
            east=self.handle.get("longitudeOfLastGridPointInDegrees"),
        )
        backend.plot_grib(self.path, self.handle.get("offset"))

    @call_counter
    def to_numpy(self, reshape=True, dtype=None):
        values = self.values
        if reshape:
            values = values.reshape(self.shape)
        if dtype is not None:
            values = values.astype(dtype)
        return values

    def __repr__(self):
        return "GribField(%s,%s,%s,%s,%s,%s)" % (
            self.handle.get("shortName"),
            self.handle.get("levelist"),
            self.handle.get("date"),
            self.handle.get("time"),
            self.handle.get("step"),
            self.handle.get("number"),
        )

    def _grid_definition(self):
        return dict(
            north=self.handle.get("latitudeOfFirstGridPointInDegrees"),
            south=self.handle.get("latitudeOfLastGridPointInDegrees"),
            west=self.handle.get("longitudeOfFirstGridPointInDegrees"),
            east=self.handle.get("longitudeOfLastGridPointInDegrees"),
            south_north_increment=self.handle.get("jDirectionIncrementInDegrees"),
            west_east_increment=self.handle.get("iDirectionIncrementInDegrees"),
        )

    def field_metadata(self):
        m = self._grid_definition()

        for n in (
            "shortName",
            "units",
            "paramId",
            "level",
            "typeOfLevel",
            "marsClass",
            "marsStream",
            "marsType",
            "number",
            "stepRange",
            "param",
            "long_name",
            "standard_name",
            "levelist",
        ):
            p = self.handle.get(n)
            if p is not None:
                m[n] = str(p)
        m["shape"] = self.shape
        return m

    @property
    def resolution(self):
        grid_type = self["gridType"]

        if grid_type == "reduced_gg":
            return self["gridName"]

        if grid_type == "regular_ll":
            x = self["DxInDegrees"]
            y = self["DyInDegrees"]
            assert x == y, (x, y)
            return x

        raise ValueError(f"Unknown gridType={grid_type}")

    def datetime(self):
        date = self.handle.get("date")
        time = self.handle.get("time")
        return datetime.datetime(
            date // 10000,
            date % 10000 // 100,
            date % 100,
            time // 100,
            time % 100,
        )

    def valid_datetime(self):
        date = self.handle.get("validityDate")
        time = self.handle.get("validityTime")
        return datetime.datetime(
            date // 10000,
            date % 10000 // 100,
            date % 100,
            time // 100,
            time % 100,
        )

    def to_datetime_list(self):
        return [self.valid_datetime()]

    def to_bounding_box(self):
        return BoundingBox(
            north=self.handle.get("latitudeOfFirstGridPointInDegrees"),
            south=self.handle.get("latitudeOfLastGridPointInDegrees"),
            west=self.handle.get("longitudeOfFirstGridPointInDegrees"),
            east=self.handle.get("longitudeOfLastGridPointInDegrees"),
        )

    def _attributes(self, names):
        result = {}
        for name in names:
            result[name] = self.handle.get(name)
        return result

    def _get(self, name):
        """Private, for testing only"""
        # paramId is renamed as param to get rid of the
        # additional '.128' (in climetlab/scripts/grib.py)
        if name == "param":
            name = "paramId"
        return self.handle.get(name)

    def metadata(self, name):
        if name == DATETIME:
            date = self.metadata("validityDate")
            time = self.metadata("validityTime")
            return datetime.datetime(
                date // 10000,
                date % 10000 // 100,
                date % 100,
                time // 100,
                time % 100,
            ).isoformat()

        if name == "param":
            name = "shortName"

        if name == "_param_id":
            name = "paramId"

        if name == "level" and self[name] == 0:
            return None

        return self[name]

    def __getitem__(self, name):
        """For cfgrib"""

        if name not in self._cache:
            proc = self.handle.get
            if ":" in name:
                try:
                    name, kind = name.split(":")
                    proc = dict(
                        str=self.handle.get_string,
                        int=self.handle.get_long,
                        float=self.handle.get_double,
                    )[kind]
                except Exception:
                    LOG.exception(f"Unsupported kind '{kind}'")
                    raise ValueError(f"Unsupported kind '{kind}'")

            self._cache[name] = proc(name)

        return self._cache[name]

    @property
    def data(self):
        return self.handle.get_data()

    def as_mars(self, param="shortName"):
        return self.handle.as_mars(param)

    def write(self, f):
        f.write(self.handle.read_bytes(self._offset, self._length))

    def plot_numpy(self, backend, array):
        if self.handle.get("gridType") == "regular_ll":
            metadata = self.field_metadata()

            backend.bounding_box(
                north=metadata["north"],
                south=metadata["south"],
                west=metadata["west"],
                east=metadata["east"],
            )

            backend.plot_numpy(
                array.reshape(metadata.get("shape", self.shape)),
                metadata=metadata,
            )
            return

        # Non-regular field
        tmp = backend.temporary_file(".grib")
        clone = self.handle.clone()
        clone.set_values(array)
        clone.save(tmp)
        GribField(tmp, 0, self._length).plot_map(backend)

    def iterate_grid_points(self):
        latlon = self.handle.get("latitudeLongitudeValues")
        yield from zip(islice(latlon, 0, None, 3), islice(latlon, 1, None, 3))

    def grid_points(self):
        import numpy as np

        data = self.data
        lat = np.array([d["lat"] for d in data])
        lon = np.array([d["lon"] for d in data])

        return lat, lon
